"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObjectNodeService = exports.ParentNodeType = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const repositories_1 = require("../repositories");
const application_error_1 = require("./../helper/application-error");
const application_service_1 = require("./application.service");
const content_entity_service_1 = require("./content-entity.service");
const object_type_service_1 = require("./object-type.service");
var ParentNodeType;
(function (ParentNodeType) {
    ParentNodeType["node"] = "node";
    ParentNodeType["tree"] = "tree";
    ParentNodeType["owner"] = "owner";
    ParentNodeType["namespace"] = "namespace";
})(ParentNodeType = exports.ParentNodeType || (exports.ParentNodeType = {}));
let ObjectNodeService = class ObjectNodeService {
    constructor(objectNodeRepository, objectTypeService, contentEntityService) {
        this.objectNodeRepository = objectNodeRepository;
        this.objectTypeService = objectTypeService;
        this.contentEntityService = contentEntityService;
    }
    searchByTreeId(treeId) {
        return this.objectNodeRepository.find({
            where: { parentTreeId: treeId },
        });
    }
    searchByTreeIds(treeIds) {
        return this.objectNodeRepository.find({
            where: { parentTreeId: { inq: treeIds } },
        });
    }
    searchByParentIdAndObjectTypeId(parentNodeId, objectTypeId) {
        return this.objectNodeRepository.find({
            where: { parentNodeId, objectTypeId },
        });
    }
    searchByParentIdsAndObjectTypeId(parentNodeIds, objectTypeId) {
        return this.objectNodeRepository.find({
            where: { parentNodeId: { inq: parentNodeIds }, objectTypeId },
        });
    }
    async searchOwner(ownerType, ownerName) {
        const ownerObjectType = await this.objectTypeService.searchByName(ownerType);
        if (!ownerObjectType) {
            throw application_error_1.ApplicationError.notFound({ ownerType });
        }
        const nodes = await this.objectNodeRepository.find({
            where: { objectTypeId: ownerObjectType.id, name: ownerName, owner: true },
        });
        if (1 < nodes.length) {
            throw application_error_1.ApplicationError.tooMany({ owner: ownerType, ownerName: ownerName });
        }
        return 1 === nodes.length ? nodes[0] : null;
    }
    async searchNamespace(ownerType, ownerName, namespaceType, namespaceName) {
        const owner = await this.searchOwner(ownerType, ownerName);
        if (!owner) {
            throw application_error_1.ApplicationError.notFound({ owner: ownerType, ownerNale: ownerName });
        }
        const namespaceObjectType = await this.objectTypeService.searchByName(namespaceType);
        if (!namespaceObjectType) {
            throw application_error_1.ApplicationError.notFound({
                owner: ownerType,
                ownerName: ownerName,
                namespace: namespaceType,
                namespaceName: namespaceName,
            });
        }
        const nodes = await this.objectNodeRepository.find({
            where: {
                objectTypeId: namespaceObjectType.id,
                name: namespaceName,
                parentOwnerId: owner.id,
                namespace: true,
            },
        });
        if (1 < nodes.length) {
            throw application_error_1.ApplicationError.tooMany({
                namespace: namespaceType,
                namespaceName: namespaceName,
                owner: ownerType,
                ownerName: ownerName,
            });
        }
        return 1 === nodes.length ? nodes[0] : null;
    }
    async getOrCreateChildren(parentId, objectTypeId, defaultValue, min = 1) {
        const children = await this.searchByParentIdAndObjectTypeId(parentId, objectTypeId);
        if (defaultValue && children.length < min) {
            while (children.length < min) {
                children.push(await this.add(lodash_1.default.merge({}, defaultValue, {
                    parentNodeId: parentId,
                    objectTypeId: objectTypeId,
                }), new application_service_1.CurrentContext()));
            }
        }
        return children;
    }
    async searchTreeNode(ownerType, ownerName, namespaceType, namespaceName, treeType, treeName) {
        const namespace = await this.searchNamespace(ownerType, ownerName, namespaceType, namespaceName);
        if (!namespace) {
            throw application_error_1.ApplicationError.notFound({
                namespace: namespaceType,
                namespaceType,
            });
        }
        const treeObjectType = await this.objectTypeService.searchByName(treeType);
        if (!treeObjectType) {
            throw application_error_1.ApplicationError.notFound({
                namespace: namespaceType,
                namespaceType,
                treeType,
            });
        }
        const nodes = await this.objectNodeRepository.find({
            where: {
                objectTypeId: treeObjectType.id,
                name: treeName,
                parentNamepaceId: namespace.id,
                tree: true,
            },
        });
        if (1 < nodes.length) {
            throw application_error_1.ApplicationError.tooMany({
                namespace: namespaceType,
                namespaceName: namespaceName,
                owner: ownerType,
                ownerName: ownerName,
            });
        }
        return 1 === nodes.length ? nodes[0] : null;
    }
    searchById(id, filter, options) {
        return this.objectNodeRepository.findById(id, filter, options);
    }
    async checkNameAvailibility(objectNode, name) {
        if (!objectNode.owner && !objectNode.namespace && !objectNode.tree) {
            return;
        }
        if (objectNode.tree) {
            const otherTrees = await this.objectNodeRepository.find({
                where: {
                    parentNamespaceId: objectNode.parentNamespaceId,
                    objectTypeId: objectNode.objectTypeId,
                    name: name,
                    tree: true,
                },
            });
            if (otherTrees && 0 < otherTrees.length) {
                throw application_error_1.ApplicationError.conflict({ name: name });
            }
        }
        if (objectNode.namespace) {
            const otherNamespaces = await this.objectNodeRepository.find({
                where: {
                    parentOwnerId: objectNode.parentOwnerId,
                    objectTypeId: objectNode.objectTypeId,
                    name: name,
                    namespace: true,
                },
            });
            if (otherNamespaces && 0 < otherNamespaces.length) {
                throw application_error_1.ApplicationError.conflict({ name: name });
            }
        }
        if (objectNode.owner) {
            const otherOwners = await this.objectNodeRepository.find({
                where: {
                    objectTypeId: objectNode.objectTypeId,
                    name: name,
                    owner: true,
                },
            });
            if (otherOwners && 0 < otherOwners.length) {
                throw application_error_1.ApplicationError.conflict({ name: name });
            }
        }
    }
    async checkBrothersCondition(objectNode, nodeContext) {
        const objectSubType = nodeContext.objectSubType.value;
        if (objectSubType.mandatories) {
            for (const brotherTypeId of objectSubType.mandatories) {
                if (0 ===
                    (await this.searchByParentIdAndObjectTypeId(objectNode.parentNodeId, brotherTypeId)).length) {
                    throw application_error_1.ApplicationError.missing({ objectType: brotherTypeId });
                }
            }
        }
        if (objectSubType.exclusions) {
            for (const brotherTypeId of objectSubType.exclusions) {
                if (0 <
                    (await this.searchByParentIdAndObjectTypeId(objectNode.parentNodeId, brotherTypeId)).length) {
                    throw application_error_1.ApplicationError.incompatible({ objectType: brotherTypeId });
                }
            }
        }
        if (objectSubType.max &&
            (await this.searchByParentIdAndObjectTypeId(objectNode.parentNodeId, objectNode.objectTypeId)).length >= objectSubType.max) {
            throw application_error_1.ApplicationError.tooMany({ objectType: objectNode.objectTypeId });
        }
    }
    async checkSubTypesCondition(objectNode, nodeContext) {
        const objectType = nodeContext.objectType.value;
        if (objectType === null || objectType === void 0 ? void 0 : objectType.objectSubTypes) {
            for (const objectSubType of objectType.objectSubTypes) {
                if (objectSubType.min && 0 < objectSubType.min) {
                    for (let i = 0; i < objectSubType.min; i++) {
                        try {
                            await this.add({
                                name: objectSubType.name,
                                objectTypeId: objectSubType.subObjectTypeId,
                                parentNodeId: objectNode.id,
                            }, application_service_1.CurrentContext.get({
                                nodeContext: {
                                    parentType: new application_service_1.ExpectedValue(objectType),
                                    parent: new application_service_1.ExpectedValue(objectNode),
                                    objectSubType: new application_service_1.ExpectedValue(objectSubType),
                                },
                            }));
                        }
                        catch (error) {
                            console.log(error);
                        }
                    }
                }
            }
        }
    }
    async add(objectNode, ctx, byPassCheck = false) {
        const nodeContext = ctx.nodeContext;
        //let objectNode = _.clone(objectNodePosted);
        let objectNodeForUpdate = objectNode;
        let objectType = null;
        if (!byPassCheck) {
            if (!objectNode.name) {
                throw application_error_1.ApplicationError.missingParameter('name');
            }
            if (!objectNode.objectTypeId) {
                throw application_error_1.ApplicationError.missingParameter('objectTypeId');
            }
            if (!objectNode.parentNodeId) {
                throw application_error_1.ApplicationError.missingParameter('parentNodeId');
            }
            const parent = await nodeContext.parent.getOrSetValue(async () => {
                return this.searchById(objectNode.parentNodeId);
            });
            if (!parent) {
                throw application_error_1.ApplicationError.notFound({ parent: objectNode.parentNodeId });
            }
            const parentType = await nodeContext.parentType.getOrSetValue(async () => {
                return this.objectTypeService.searchById(parent.objectTypeId);
            });
            if (!parentType) {
                throw application_error_1.ApplicationError.corruptedData({
                    parent: objectNode.parentNodeId,
                    parentType: parent.objectTypeId,
                });
            }
            const objectSubType = await nodeContext.objectSubType.getOrSetValue(async () => {
                return lodash_1.default.find(parentType.objectSubTypes, (subType) => {
                    return subType.subObjectTypeId === objectNode.objectTypeId;
                });
            });
            if (!objectSubType) {
                throw application_error_1.ApplicationError.unauthorizedValue({
                    objectType: objectNode.objectTypeId,
                });
            }
            await this.checkBrothersCondition(objectNode, nodeContext);
            objectType = await nodeContext.objectType.getOrSetValue(async () => {
                return this.objectTypeService.searchById(objectNode.objectTypeId);
            });
            if (!objectType) {
                throw application_error_1.ApplicationError.notFound({ objectType: objectNode.objectTypeId });
            }
            objectNode.parentACLId =
                !parent.acl && parent.parentACLId ? parent.parentACLId : parent.id;
            objectNode.aclList = !parent.aclList
                ? [parent.id]
                : objectNode.parentACLId === parent.id
                    ? lodash_1.default.concat(parent.aclList, [objectNode.parentACLId])
                    : parent.aclList;
            objectNode.parentOwnerId =
                !parent.owner && parent.parentOwnerId
                    ? parent.parentOwnerId
                    : parent.id;
            objectNode.parentTreeId =
                !parent.tree && parent.parentTreeId ? parent.parentTreeId : parent.id;
            objectNode.parentNamespaceId =
                !parent.namespace && parent.parentNamespaceId
                    ? parent.parentNamespaceId
                    : parent.id;
            objectNode.owner = objectSubType.owner;
            objectNode.acl = !!objectNode.owner || !!objectSubType.acl;
            objectNode.namespace = !!objectNode.owner || !!objectSubType.namespace;
            objectNode.tree = !!objectNode.namespace || !!objectSubType.tree;
            await this.checkNameAvailibility(objectNode, objectNode.name);
            objectNodeForUpdate = lodash_1.default.pick(objectNode, this.getPropertiesKeys(objectType, [
                'name',
                'objectTypeId',
                'parentNodeId',
                'parentACLId',
                'aclList',
                'parentOwnerId',
                'parentTreeId',
                'parentNamespaceId',
                'owner',
                'acl',
                'namespace',
                'tree',
            ]));
        }
        const result = await this.objectNodeRepository.create(objectNodeForUpdate);
        const changes = await this.contentEntityService.manageContent(objectType === null || objectType === void 0 ? void 0 : objectType.contentType, result, objectNode);
        if (changes) {
            await this.objectNodeRepository.updateById(result.id, result);
        }
        await this.checkSubTypesCondition(result, nodeContext);
        await this.contentEntityService.addTransientContent(objectType === null || objectType === void 0 ? void 0 : objectType.contentType, result);
        return result;
    }
    getPropertiesKeys(objectType, objectKeys = ['name']) {
        var _a;
        let finalObjectKeys = objectKeys;
        if ((_a = objectType.definition) === null || _a === void 0 ? void 0 : _a.properties) {
            finalObjectKeys = objectKeys.concat(Object.keys(objectType.definition.properties));
        }
        return finalObjectKeys;
    }
    async modifyById(id, objectNode, ctx) {
        const node = await this.objectNodeRepository.findById(id);
        if (!node) {
            throw application_error_1.ApplicationError.notFound({ object: id });
        }
        if ('name' in objectNode && node.name !== objectNode.name) {
            if (!objectNode.name) {
                throw application_error_1.ApplicationError.missingParameter('name');
            }
            await this.checkNameAvailibility(node, objectNode.name);
        }
        const objectType = await this.objectTypeService.searchById(node.objectTypeId);
        if (!objectType) {
            throw application_error_1.ApplicationError.corruptedData({
                object: id,
                objectType: node.objectTypeId,
            });
        }
        await this.objectNodeRepository.updateById(id, lodash_1.default.pick(objectNode, this.getPropertiesKeys(objectType)));
        const result = await this.objectNodeRepository.findById(id);
        const changes = await this.contentEntityService.manageContent(objectType === null || objectType === void 0 ? void 0 : objectType.contentType, result, objectNode);
        if (changes) {
            await this.objectNodeRepository.updateById(result.id, result);
        }
        await this.contentEntityService.addTransientContent(objectType === null || objectType === void 0 ? void 0 : objectType.contentType, result);
        return result;
    }
    getParentIdKey(parentType) {
        return ('parent' +
            parentType.charAt(0).toUpperCase() +
            parentType.substr(1) +
            'Id');
    }
    async removeByParent(parentType, id) {
        const parentIdKey = this.getParentIdKey(parentType);
        const whereClause = lodash_1.default.merge({
            [parentIdKey]: id,
        }, 'node' === parentType ? {} : { [parentType]: true });
        const subChildsNodes = await this.objectNodeRepository.find({
            where: whereClause,
        });
        for (const subChildNode of subChildsNodes) {
            await this.removeByParent(parentType, subChildNode.id);
        }
        await this.objectNodeRepository.deleteAll({
            [parentIdKey]: id,
        });
    }
    async removeById(id, ctx) {
        const node = await this.objectNodeRepository.findById(id);
        if (!node) {
            throw application_error_1.ApplicationError.notFound({ object: id });
        }
        if (node.owner) {
            await this.removeByParent(ParentNodeType.owner, node.id);
        }
        else if (node.namespace) {
            await this.removeByParent(ParentNodeType.namespace, node.id);
        }
        else if (node.tree) {
            await this.removeByParent(ParentNodeType.tree, node.id);
        }
        else {
            await this.removeByParent(ParentNodeType.node, node.id);
        }
        await this.objectNodeRepository.deleteById(id);
        return;
    }
};
ObjectNodeService = tslib_1.__decorate([
    core_1.bind({ scope: core_1.BindingScope.SINGLETON }),
    tslib_1.__param(0, repository_1.repository(repositories_1.ObjectNodeRepository)),
    tslib_1.__param(1, core_1.service(object_type_service_1.ObjectTypeService)),
    tslib_1.__param(2, core_1.service(content_entity_service_1.ContentEntityService)),
    tslib_1.__metadata("design:paramtypes", [repositories_1.ObjectNodeRepository,
        object_type_service_1.ObjectTypeService,
        content_entity_service_1.ContentEntityService])
], ObjectNodeService);
exports.ObjectNodeService = ObjectNodeService;
//# sourceMappingURL=object-node.service.js.map