"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObjectTreeService = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@loopback/core");
const _ = tslib_1.__importStar(require("lodash"));
const object_tree_model_1 = require("../../models/object-tree.model");
const application_service_1 = require("../application.service");
const content_entity_service_1 = require("../content-entity.service");
const object_node_service_1 = require("../object-node.service");
const object_type_service_1 = require("../object-type.service");
const application_error_1 = require("./../../helper/application-error");
const object_tree_init_1 = require("./object-tree.init");
let ObjectTreeService = class ObjectTreeService {
    constructor(objectNodeService, objectTypeService, contentEntityService, appCtx) {
        this.objectNodeService = objectNodeService;
        this.objectTypeService = objectTypeService;
        this.contentEntityService = contentEntityService;
        this.appCtx = appCtx;
        this.init = new object_tree_init_1.ObjectTreeInit(objectNodeService, objectTypeService, contentEntityService, appCtx);
    }
    get ready() {
        return this.init.ready;
    }
    /*
     * Add service methods here
     */
    async getOwnerTreeNodes(ownerType, ownerName, ctx) {
        await this.ready;
        const tree = await ctx.treeContext.treeNode.getOrSetValue(async () => {
            return this.objectNodeService.searchOwner(ownerType, ownerName);
        });
        if (!tree) {
            throw application_error_1.ApplicationError.notFound({ owner: ownerType, ownerName: ownerName });
        }
        return _.concat(tree, await this.objectNodeService.searchByTreeId(tree.id));
    }
    async buildTreeFromNodes(treeNode, availableNodes, ctx) {
        return new object_tree_model_1.ObjectTree(treeNode).init(availableNodes, this.contentEntityService, await this.objectTypeService.getAll(ctx));
    }
    async getOwnerTree(ownerType, ownerName, ctx) {
        const objectNodes = await this.getOwnerTreeNodes(ownerType, ownerName, ctx);
        return this.buildTreeFromNodes(objectNodes[0], objectNodes, ctx);
    }
    async getTreeNodes(ownerType, ownerName, namespaceType, namespaceName, treeType, treeName, ctx) {
        await this.ready;
        const tree = await ctx.treeContext.treeNode.getOrSetValue(async () => {
            return this.objectNodeService.searchTreeNode(ownerType, ownerName, namespaceType, namespaceName, treeType, treeName);
        });
        if (!tree) {
            throw application_error_1.ApplicationError.notFound({
                tree: treeType,
                treeName: treeName,
                namespace: namespaceType,
                namespaceName: namespaceName,
                owner: ownerType,
                ownerName: ownerName,
            });
        }
        return _.concat(tree, await this.objectNodeService.searchByTreeId(tree.id));
    }
    async getTree(ownerType, ownerName, namespaceType, namespaceName, treeType, treeName, ctx) {
        const objectNodes = await this.getTreeNodes(ownerType, ownerName, namespaceType, namespaceName, treeType, treeName, ctx);
        return this.buildTreeFromNodes(objectNodes[0], objectNodes, ctx);
    }
    async getNamespaceNodes(ownerType, ownerName, namespaceType, namespaceName, ctx) {
        await this.ready;
        const namespace = await ctx.treeContext.treeNode.getOrSetValue(async () => {
            return this.objectNodeService.searchNamespace(ownerType, ownerName, namespaceType, namespaceName);
        });
        if (!namespace) {
            throw application_error_1.ApplicationError.notFound({
                namespace: namespaceType,
                namespaceName: namespaceName,
                owner: ownerType,
                ownerName: ownerName,
            });
        }
        return _.concat(namespace, await this.objectNodeService.searchByTreeId(namespace.id));
    }
    async gettNamespaceTree(ownerType, ownerName, namespaceType, namespaceName, ctx) {
        const objectNodes = await this.getNamespaceNodes(ownerType, ownerName, namespaceType, namespaceName, ctx);
        return this.buildTreeFromNodes(objectNodes[0], objectNodes, ctx);
    }
    async loadChildrenNodes(treeId, ctx) {
        const root = await ctx.treeContext.treeNode.getOrSetValue(async () => {
            return this.objectNodeService.searchById(treeId);
        });
        if (!root) {
            throw application_error_1.ApplicationError.notFound({
                tree: treeId,
            });
        }
        return _.concat(root, await this.objectNodeService.searchByTreeId(treeId));
    }
    async loadTree(treeId, ctx) {
        const objectNodes = await this.loadChildrenNodes(treeId, ctx);
        return this.buildTreeFromNodes(objectNodes[0], objectNodes, ctx);
    }
    isInTree(tree, node) {
        if (node.id === tree.treeNode.id) {
            return true;
        }
        for (const child of tree.children) {
            if (this.isInTree(child, node)) {
                return true;
            }
        }
        return false;
    }
    getChildOfType(tree, objectTypeId) {
        for (const child of tree.children) {
            if (child.treeNode.objectTypeId === objectTypeId) {
                return child;
            }
        }
        return null;
    }
};
ObjectTreeService = tslib_1.__decorate([
    core_1.bind({ scope: core_1.BindingScope.SINGLETON }),
    tslib_1.__param(0, core_1.service(object_node_service_1.ObjectNodeService)),
    tslib_1.__param(1, core_1.service(object_type_service_1.ObjectTypeService)),
    tslib_1.__param(2, core_1.service(content_entity_service_1.ContentEntityService)),
    tslib_1.__param(3, core_1.service(application_service_1.ApplicationService)),
    tslib_1.__metadata("design:paramtypes", [object_node_service_1.ObjectNodeService,
        object_type_service_1.ObjectTypeService,
        content_entity_service_1.ContentEntityService,
        application_service_1.ApplicationService])
], ObjectTreeService);
exports.ObjectTreeService = ObjectTreeService;
//# sourceMappingURL=object-tree.service.js.map