"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObjectTypeService = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const repositories_1 = require("../repositories");
const application_error_1 = require("./../helper/application-error");
const object_sub_type_repository_1 = require("./../repositories/object-sub-type.repository");
const application_service_1 = require("./application.service");
const content_entity_service_1 = require("./content-entity.service");
const defaultObjectTypeFilter = {
    order: ['name'],
    fields: {
        definition: true,
        id: true,
        name: true,
        contentType: true,
        uri: true,
    },
    include: [
        {
            relation: 'objectSubTypes',
            scope: {
                order: ['index'],
            },
        },
    ],
};
let ObjectTypeService = class ObjectTypeService {
    constructor(objectTypeRepository, objectSubTypeRepository, contentEntityService, appCtx) {
        this.objectTypeRepository = objectTypeRepository;
        this.objectSubTypeRepository = objectSubTypeRepository;
        this.contentEntityService = contentEntityService;
        this.appCtx = appCtx;
    }
    async add(objectType, ctx) {
        delete objectType.contentDefinition;
        const newType = await this.objectTypeRepository.create(objectType);
        return newType;
    }
    async searchByName(name) {
        const objectTypes = await this.objectTypeRepository.find({
            where: { name: name },
        });
        if (1 < objectTypes.length) {
            throw application_error_1.ApplicationError.tooMany({ objectType: name });
        }
        if (!!objectTypes && 0 < objectTypes.length) {
            const objectType = objectTypes[0];
            if (objectType) {
                objectType.contentDefinition = await this.contentEntityService.getContentDefinition(objectType.contentType);
            }
            return objectType;
        }
        return null;
    }
    get filterOrder() {
        return lodash_1.default.cloneDeep(defaultObjectTypeFilter.order);
    }
    get filterFields() {
        return lodash_1.default.cloneDeep(defaultObjectTypeFilter.fields);
    }
    get filterInclude() {
        return lodash_1.default.cloneDeep(defaultObjectTypeFilter.include);
    }
    async searchById(id, filter, options) {
        if (!filter) {
            filter = {};
        }
        if (!filter.fields) {
            filter.fields = this.filterFields;
        }
        filter.include = this.filterInclude;
        const objectType = await this.objectTypeRepository.findById(id, filter, options);
        if (objectType) {
            objectType.contentDefinition = await this.contentEntityService.getContentDefinition(objectType.contentType);
        }
        return objectType;
    }
    async search(ctx) {
        const filter = {
            order: this.filterOrder,
            fields: this.filterFields,
            include: this.filterInclude,
        };
        const objectTypes = await this.objectTypeRepository.find(filter);
        for (const objectType of objectTypes) {
            objectType.contentDefinition = await this.contentEntityService.getContentDefinition(objectType.contentType);
        }
        return objectTypes;
    }
    async getAll(ctx) {
        return ctx.typeContext.types.getOrSetValue(async () => {
            return lodash_1.default.mapKeys(await this.objectTypeRepository.find(), (objectType, index) => {
                return objectType.id;
            });
        });
    }
    removeById(id, ctx) {
        return this.objectTypeRepository.deleteById(id);
    }
    async modifyById(id, objectType, ctx) {
        delete objectType.contentDefinition;
        await this.objectTypeRepository.updateById(id, objectType);
        return this.objectTypeRepository.findById(id);
    }
    async createSubType(id, objectSubType) {
        return this.objectTypeRepository.objectSubTypes(id).create(objectSubType);
    }
    async modifySubTypeById(objectTypeId, id, objectSubType) {
        const where = {
            id: id,
        };
        await this.objectTypeRepository
            .objectSubTypes(objectTypeId)
            .patch(objectSubType, where);
        return this.objectSubTypeRepository.findById(id);
    }
    async removeSubTypeById(objectTypeId, id) {
        const where = {
            id: id,
        };
        return this.objectTypeRepository.objectSubTypes(objectTypeId).delete(where);
    }
    async getOrCreateObjectSubType(objectTypeId, subObjectTypeId, defaultValue) {
        const where = {
            subObjectTypeId: subObjectTypeId,
        };
        const subTypes = await this.objectTypeRepository
            .objectSubTypes(objectTypeId)
            .find({ where });
        if (!subTypes || 0 === subTypes.length) {
            if (!defaultValue) {
                return null;
            }
            return this.createSubType(objectTypeId, lodash_1.default.merge({}, defaultValue, { subObjectTypeId }));
        }
        if (1 < subTypes.length) {
            throw application_error_1.ApplicationError.tooMany({
                objectType: objectTypeId,
                subObjectType: subObjectTypeId,
            });
        }
        return subTypes[0];
    }
};
ObjectTypeService = tslib_1.__decorate([
    core_1.bind({ scope: core_1.BindingScope.SINGLETON }),
    tslib_1.__param(0, repository_1.repository(repositories_1.ObjectTypeRepository)),
    tslib_1.__param(1, repository_1.repository(object_sub_type_repository_1.ObjectSubTypeRepository)),
    tslib_1.__param(2, core_1.service(content_entity_service_1.ContentEntityService)),
    tslib_1.__param(3, core_1.service(application_service_1.ApplicationService)),
    tslib_1.__metadata("design:paramtypes", [repositories_1.ObjectTypeRepository,
        object_sub_type_repository_1.ObjectSubTypeRepository,
        content_entity_service_1.ContentEntityService,
        application_service_1.ApplicationService])
], ObjectTypeService);
exports.ObjectTypeService = ObjectTypeService;
//# sourceMappingURL=object-type.service.js.map